/** @file */

#ifndef OPENMESH_PYTHON_MESHTYPES_HH
#define OPENMESH_PYTHON_MESHTYPES_HH

#define OM_STATIC_BUILD

#include "Utilities.hh"
#include "TypeConversions.hh"

#include <OpenMesh/Core/IO/MeshIO.hh>

#ifdef OPENMESH_PYTHON_OPENFLIPPER_TYPES
#include <ObjectTypes/TriangleMesh/TriangleMeshTypes.hh>
#include <ObjectTypes/PolyMesh/PolyMeshTypes.hh>
#else
#include <OpenMesh/Core/Mesh/TriMesh_ArrayKernelT.hh>
#include <OpenMesh/Core/Mesh/PolyMesh_ArrayKernelT.hh>
#endif

#include <pybind11/pybind11.h>
#include <pybind11/numpy.h>

namespace py = pybind11;

namespace OpenMeshPython {


struct MeshTraits : public OpenMesh::DefaultTraits {
	/** Use double precision points */
	typedef OpenMesh::Vec3d Point;

	/** Use double precision normals */
	typedef OpenMesh::Vec3d Normal;

	/** Use RGBA colors */
	typedef OpenMesh::Vec4f Color;

	/** Use double precision texcoords */
	typedef double TexCoord1D;
	typedef OpenMesh::Vec2d TexCoord2D;
	typedef OpenMesh::Vec3d TexCoord3D;
};


template <class Mesh>
class MeshWrapperT : public Mesh {
public:

	typedef OpenMesh::VPropHandleT<py::none> VPropHandle;
	typedef OpenMesh::HPropHandleT<py::none> HPropHandle;
	typedef OpenMesh::EPropHandleT<py::none> EPropHandle;
	typedef OpenMesh::FPropHandleT<py::none> FPropHandle;

	template <class Handle>
	py::object py_property(const std::string& _name, Handle _h) {
		#ifdef OPENMESH_PYTHON_HIDE_CPP_PROPERTIES
		const auto prop = py_prop_on_demand<Handle, HandleToPropHandle<Handle, py::none>>(_name);
		return Mesh::property(prop, _h);
		#else
		OpenMesh::BaseProperty *base_prop = nullptr;
		int idx = -1;
		if (py_get_base_property<Handle>(base_prop, idx, _name)) {
			// there is a property with the given name
			if (dynamic_cast<OpenMesh::PropertyT<py::none>*>(base_prop)) {
				HandleToPropHandle<Handle, py::none> prop(idx);
				return Mesh::property(prop, _h);
			}
			else if (dynamic_cast<OpenMesh::PropertyT<bool>*>(base_prop)) {
				HandleToPropHandle<Handle, bool> prop(idx);
				return convert_to_python(Mesh::property(prop, _h));
			}
			else if (dynamic_cast<OpenMesh::PropertyT<int>*>(base_prop)) {
				HandleToPropHandle<Handle, int> prop(idx);
				return convert_to_python(Mesh::property(prop, _h));
			}
			else if (dynamic_cast<OpenMesh::PropertyT<float>*>(base_prop)) {
				HandleToPropHandle<Handle, float> prop(idx);
				return convert_to_python(Mesh::property(prop, _h));
			}
			else if (dynamic_cast<OpenMesh::PropertyT<double>*>(base_prop)) {
				HandleToPropHandle<Handle, double> prop(idx);
				return convert_to_python(Mesh::property(prop, _h));
			}
			else if (dynamic_cast<OpenMesh::PropertyT<OpenMesh::Vec2f>*>(base_prop)) {
				HandleToPropHandle<Handle, OpenMesh::Vec2f> prop(idx);
				return vec2numpy(*this, Mesh::property(prop, _h));
			}
			else if (dynamic_cast<OpenMesh::PropertyT<OpenMesh::Vec3f>*>(base_prop)) {
				HandleToPropHandle<Handle, OpenMesh::Vec3f> prop(idx);
				return vec2numpy(*this, Mesh::property(prop, _h));
			}
			else if (dynamic_cast<OpenMesh::PropertyT<OpenMesh::Vec4f>*>(base_prop)) {
				HandleToPropHandle<Handle, OpenMesh::Vec4f> prop(idx);
				return vec2numpy(*this, Mesh::property(prop, _h));
			}
			else if (dynamic_cast<OpenMesh::PropertyT<OpenMesh::Vec2d>*>(base_prop)) {
				HandleToPropHandle<Handle, OpenMesh::Vec2d> prop(idx);
				return vec2numpy(*this, Mesh::property(prop, _h));
			}
			else if (dynamic_cast<OpenMesh::PropertyT<OpenMesh::Vec3d>*>(base_prop)) {
				HandleToPropHandle<Handle, OpenMesh::Vec3d> prop(idx);
				return vec2numpy(*this, Mesh::property(prop, _h));
			}
			else if (dynamic_cast<OpenMesh::PropertyT<OpenMesh::Vec4d>*>(base_prop)) {
				HandleToPropHandle<Handle, OpenMesh::Vec4d> prop(idx);
				return vec2numpy(*this, Mesh::property(prop, _h));
			}
			else {
				PyErr_SetString(PyExc_RuntimeError, "Unsupported C++ property type.");
				throw py::error_already_set();
			}
		}
		else {
			// there is no property with the given name
			const auto prop = py_prop_on_demand<Handle, HandleToPropHandle<Handle, py::none>>(_name);
			return Mesh::property(prop, _h);
		}
		#endif
	}

	template <class Handle>
	void py_set_property(const std::string& _name, Handle _h, py::object _val) {
		#ifdef OPENMESH_PYTHON_HIDE_CPP_PROPERTIES
		const auto prop = py_prop_on_demand<Handle, HandleToPropHandle<Handle, py::none>>(_name);
		Mesh::property(prop, _h) = _val;
		#else
		OpenMesh::BaseProperty *base_prop = nullptr;
		int idx = -1;
		if (py_get_base_property<Handle>(base_prop, idx, _name)) {
			// there is a property with the given name, so check
			// its type and covert _val to the correct c++ type
			// if necessary
			if (dynamic_cast<OpenMesh::PropertyT<py::none>*>(base_prop)) {
				HandleToPropHandle<Handle, py::none> prop(idx);
				Mesh::property(prop, _h) = _val;
			}
			else if (dynamic_cast<OpenMesh::PropertyT<bool>*>(base_prop)) {
				HandleToPropHandle<Handle, bool> prop(idx);
				Mesh::property(prop, _h) = convert_to_cpp<bool>(_val);
			}
			else if (dynamic_cast<OpenMesh::PropertyT<int>*>(base_prop)) {
				HandleToPropHandle<Handle, int> prop(idx);
				Mesh::property(prop, _h) = convert_to_cpp<int>(_val);
			}
			else if (dynamic_cast<OpenMesh::PropertyT<float>*>(base_prop)) {
				HandleToPropHandle<Handle, float> prop(idx);
				Mesh::property(prop, _h) = convert_to_cpp<float>(_val);
			}
			else if (dynamic_cast<OpenMesh::PropertyT<double>*>(base_prop)) {
				HandleToPropHandle<Handle, double> prop(idx);
				Mesh::property(prop, _h) = convert_to_cpp<double>(_val);
			}
			else if (dynamic_cast<OpenMesh::PropertyT<OpenMesh::Vec2f>*>(base_prop)) {
				HandleToPropHandle<Handle, OpenMesh::Vec2f> prop(idx);
				Mesh::property(prop, _h) = convert_to_cpp<OpenMesh::Vec2f>(_val);
			}
			else if (dynamic_cast<OpenMesh::PropertyT<OpenMesh::Vec3f>*>(base_prop)) {
				HandleToPropHandle<Handle, OpenMesh::Vec3f> prop(idx);
				Mesh::property(prop, _h) = convert_to_cpp<OpenMesh::Vec3f>(_val);
			}
			else if (dynamic_cast<OpenMesh::PropertyT<OpenMesh::Vec4f>*>(base_prop)) {
				HandleToPropHandle<Handle, OpenMesh::Vec4f> prop(idx);
				Mesh::property(prop, _h) = convert_to_cpp<OpenMesh::Vec4f>(_val);
			}
			else if (dynamic_cast<OpenMesh::PropertyT<OpenMesh::Vec2d>*>(base_prop)) {
				HandleToPropHandle<Handle, OpenMesh::Vec2d> prop(idx);
				Mesh::property(prop, _h) = convert_to_cpp<OpenMesh::Vec2d>(_val);
			}
			else if (dynamic_cast<OpenMesh::PropertyT<OpenMesh::Vec3d>*>(base_prop)) {
				HandleToPropHandle<Handle, OpenMesh::Vec3d> prop(idx);
				Mesh::property(prop, _h) = convert_to_cpp<OpenMesh::Vec3d>(_val);
			}
			else if (dynamic_cast<OpenMesh::PropertyT<OpenMesh::Vec4d>*>(base_prop)) {
				HandleToPropHandle<Handle, OpenMesh::Vec4d> prop(idx);
				Mesh::property(prop, _h) = convert_to_cpp<OpenMesh::Vec4d>(_val);
			}
			else {
				PyErr_SetString(PyExc_RuntimeError, "Unsupported C++ property type.");
				throw py::error_already_set();
			}
		}
		else {
			// there is no property with the given name
			const auto prop = py_prop_on_demand<Handle, HandleToPropHandle<Handle, py::none>>(_name);
			Mesh::property(prop, _h) = _val;
		}
		#endif
	}

	template <class Handle>
	bool py_has_property(const std::string& _name) {
		#ifdef OPENMESH_PYTHON_HIDE_CPP_PROPERTIES
		HandleToPropHandle<Handle, py::none> dummy;
		return py_get_property_handle<Handle>(dummy, _name);
		#else
		OpenMesh::BaseProperty *base_prop = nullptr;
		int idx = -1;
		return py_get_base_property<Handle>(base_prop, idx, _name);
		#endif
	}

	template <class Handle>
	void py_remove_property(const std::string& _name) {
		#ifdef OPENMESH_PYTHON_HIDE_CPP_PROPERTIES
		HandleToPropHandle<Handle, py::none> h;
		Mesh::get_property_handle(h, _name);
		Mesh::remove_property(h);
		#else
		OpenMesh::BaseProperty *base_prop = nullptr;
		int idx = -1;
		if (py_get_base_property<Handle>(base_prop, idx, _name)) {
			// this only works because OpenMesh does not check the
			// property type in remove_property()
			HandleToPropHandle<Handle, py::none> h(idx);
			Mesh::remove_property(h);
		}
		#endif
	}

	template <class Handle, class PropHandle>
	py::list py_property_generic(const std::string& _name) {
		const size_t n = py_n_items(Handle());
		const auto prop = py_prop_on_demand<Handle, PropHandle>(_name);

		py::list res;
		for (size_t i = 0; i < n; ++i) {
			res.append(Mesh::property(prop, Handle(i)));
		}
		return res;
	}

	template <class Handle, class PropHandle>
	void py_set_property_generic(const std::string& _name, py::list _list) {
		const size_t n = py_n_items(Handle());
		const auto prop = py_prop_on_demand<Handle, PropHandle>(_name);

		if (_list.size() != n) {
			PyErr_SetString(PyExc_RuntimeError, "List must have length n.");
			throw py::error_already_set();
		}
		for (size_t i = 0; i < n; ++i) {
			Mesh::property(prop, Handle(i)) = py::object(_list[i]);
		}
	}

	template <class Handle, class PropHandle>
	py::array_t<double> py_property_array(const std::string& _name) {
		const size_t n = py_n_items(Handle());
		const auto prop = py_prop_on_demand<Handle, PropHandle>(_name);

		// assume that all arrays have the same size and
		// retrieve the size of the first array
		const py::object tmp_obj = Mesh::property(prop, Handle(0));
		const auto first_arr = make_c_style(tmp_obj.cast<py::array_t<double> >());
		const size_t size = first_arr.size();

		// better check this now
		if (size == 0) {
			PyErr_SetString(PyExc_RuntimeError, "One of the arrays has size 0.");
			throw py::error_already_set();
		}

		// preserve array shape and strides
		std::vector<size_t> shape({n});
		std::vector<size_t> strides({size * sizeof(double)});
		shape.insert(shape.end(), first_arr.shape(), first_arr.shape() + first_arr.ndim());
		strides.insert(strides.end(), first_arr.strides(), first_arr.strides() + first_arr.ndim());

		// allocate memory
		double *data = new double[size * n];
		py::capsule base = free_when_done(data);

		// copy one array at a time
		for (size_t i = 0; i < n; ++i) {
			const Handle hnd(i);
			const py::object obj = Mesh::property(prop, hnd);
			const auto arr = make_c_style(obj.cast<py::array_t<double> >());
			if (arr.size() != first_arr.size()) {
				PyErr_SetString(PyExc_RuntimeError, "Array sizes do not match.");
				throw py::error_already_set();
			}
			if (arr.ndim() != first_arr.ndim()) {
				PyErr_SetString(PyExc_RuntimeError, "Array dimensions do not match.");
				throw py::error_already_set();
			}
			if (!std::equal(arr.shape(), arr.shape() + arr.ndim(), first_arr.shape())) {
				PyErr_SetString(PyExc_RuntimeError, "Array shapes do not match.");
				throw py::error_already_set();
			}
			std::copy(arr.data(), arr.data() + size, &data[size * i]);
		}

		return py::array_t<double>(shape, strides, data, base);
	}

	template <class Handle, class PropHandle>
	void py_set_property_array(const std::string& _name, py::array_t<double, py::array::c_style | py::array::forcecast> _arr) {
		const size_t n = py_n_items(Handle());
		const auto prop = py_prop_on_demand<Handle, PropHandle>(_name);

		// array cannot be empty and its shape has to be (n, ...)
		if (_arr.size() == 0 || _arr.ndim() < 1 || _arr.shape(0) < 0 || (size_t)_arr.shape(0) != n) {
			PyErr_SetString(PyExc_RuntimeError, "Array must have shape (n, ...).");
			throw py::error_already_set();
		}

		if (_arr.ndim() == 1) {
			// special case: use scalar properties
			for (size_t i = 0; i < n; ++i) {
				Mesh::property(prop, Handle(i)) = py::float_(_arr.at(i));
			}
		}
		else {
			// copy one array at a time
			const size_t size = _arr.strides(0) / sizeof(double);
			for (size_t i = 0; i < n; ++i) {
				double *data = new double[size];
				std::copy(_arr.data(i), _arr.data(i) + size, data);
				const std::vector<size_t> shape(_arr.shape() + 1, _arr.shape() + _arr.ndim());
				const std::vector<size_t> strides(_arr.strides() + 1, _arr.strides() + _arr.ndim());
				py::capsule base = free_when_done(data);
				py::array_t<double> tmp(shape, strides, data, base);
				Mesh::property(prop, Handle(i)) = tmp;
			}
		}
	}

	template <class Handle, class PropHandle>
	void py_copy_property(const std::string& _name, Handle _from, Handle _to) {
		auto prop = py_prop_on_demand<Handle, PropHandle>(_name);
		Mesh::copy_property(prop, _from, _to);
	}

	py::object py_copy() {
		return py::cast(MeshWrapperT(*this));
	}

	py::object py_deepcopy(py::dict _memo) {
		#if PY_MAJOR_VERSION < 3
		py::object id = py::module::import("__builtin__").attr("id");
		#else
		py::object id = py::module::import("builtins").attr("id");
		#endif
		py::object deepcopy = py::module::import("copy").attr("deepcopy");

		MeshWrapperT *copy = new MeshWrapperT(*this);
		py::object copy_pyobj = py::cast(copy, py::return_value_policy::take_ownership);
		_memo[id(py::cast(this))] = copy_pyobj;

		py_deepcopy_prop<OpenMesh::VertexHandle>(copy, deepcopy, _memo);
		py_deepcopy_prop<OpenMesh::HalfedgeHandle>(copy, deepcopy, _memo);
		py_deepcopy_prop<OpenMesh::EdgeHandle>(copy, deepcopy, _memo);
		py_deepcopy_prop<OpenMesh::FaceHandle>(copy, deepcopy, _memo);

		return copy_pyobj;
	}

	template <class Handle>
	void py_add_cpp_property(const std::string& _name, const std::string& _type) {
		if (py_has_property<Handle>(_name)) {
			PyErr_SetString(PyExc_RuntimeError, "A property with the given name already exists.");
			throw py::error_already_set();
		}
		if (_type == "bool") {
			HandleToPropHandle<Handle, bool> prop;
			Mesh::add_property(prop, _name);
		}
		else if (_type == "int") {
			HandleToPropHandle<Handle, int> prop;
			Mesh::add_property(prop, _name);
		}
		else if (_type == "float") {
			HandleToPropHandle<Handle, float> prop;
			Mesh::add_property(prop, _name);
		}
		else if (_type == "double") {
			HandleToPropHandle<Handle, double> prop;
			Mesh::add_property(prop, _name);
		}
		else if (_type == "vec2f") {
			HandleToPropHandle<Handle, OpenMesh::Vec2f> prop;
			Mesh::add_property(prop, _name);
		}
		else if (_type == "vec3f") {
			HandleToPropHandle<Handle, OpenMesh::Vec3f> prop;
			Mesh::add_property(prop, _name);
		}
		else if (_type == "vec4f") {
			HandleToPropHandle<Handle, OpenMesh::Vec4f> prop;
			Mesh::add_property(prop, _name);
		}
		else if (_type == "vec2d") {
			HandleToPropHandle<Handle, OpenMesh::Vec2d> prop;
			Mesh::add_property(prop, _name);
		}
		else if (_type == "vec3d") {
			HandleToPropHandle<Handle, OpenMesh::Vec3d> prop;
			Mesh::add_property(prop, _name);
		}
		else if (_type == "vec4d") {
			HandleToPropHandle<Handle, OpenMesh::Vec4d> prop;
			Mesh::add_property(prop, _name);
		}
		else {
			PyErr_SetString(PyExc_RuntimeError, "Unsupported C++ property type.");
			throw py::error_already_set();
		}
	}

	size_t py_n_items(OpenMesh::VertexHandle) const { return Mesh::n_vertices(); }
	size_t py_n_items(OpenMesh::HalfedgeHandle) const { return Mesh::n_halfedges(); }
	size_t py_n_items(OpenMesh::EdgeHandle) const { return Mesh::n_edges(); }
	size_t py_n_items(OpenMesh::FaceHandle) const { return Mesh::n_faces(); }

	size_t py_has_status(OpenMesh::VertexHandle) const { return Mesh::has_vertex_status(); }
	size_t py_has_status(OpenMesh::HalfedgeHandle) const { return Mesh::has_halfedge_status(); }
	size_t py_has_status(OpenMesh::EdgeHandle) const { return Mesh::has_edge_status(); }
	size_t py_has_status(OpenMesh::FaceHandle) const { return Mesh::has_face_status(); }

private:

	template <class Handle>
	void py_deepcopy_prop(MeshWrapperT *_copy, py::object _copyfunc, py::dict _memo) {
		using PropHandle = HandleToPropHandle<Handle, py::none>;
		const auto enditer = this->end(Handle());
		for (auto iter = this->begin(Handle()); iter != enditer; ++iter){

			// only select properties owned by python (all properties with type "py::none")
			const OpenMesh::BaseProperty* baseProp = *iter;
			const auto* srcProp = dynamic_cast< const OpenMesh::PropertyT<py::none>* >(baseProp);
			if (!srcProp)
				continue; //not a python prop -> dont copy

			const PropHandle targetProp = _copy->py_prop_on_demand<Handle, PropHandle>(srcProp->name());

			for (size_t i = 0; i < srcProp->n_elements(); ++i) {
				_copy->property(targetProp, Handle(i)) = _copyfunc((*srcProp)[i], _memo);
			}
		}
	}

	template <class Handle, class PropHandle>
	bool py_get_property_handle(PropHandle &_out, const std::string& _name) {
		#ifdef OPENMESH_PYTHON_HIDE_CPP_PROPERTIES
		// Enable type checking for OpenMesh Properties.
		// In OpenMesh it is disabled due to a problem with the shared linkage using libc++
		// We link the static library and therefore, OpenMesh-Python is not affected by this
		// This function enables type checking, if it is disabled in OpenMesh (per default, in Relase mode)
		// Use normal OpenMesh property_handle function, when type checking is not disabled (debug more or hopefully in the future)
		int idx = -1;
		const auto it = std::find_if(this->begin(Handle()), this->end(Handle()), [&_name, &idx](OpenMesh::BaseProperty* p)
		{
			++idx;
			return ((p != nullptr) && (p->name() == _name) && (dynamic_cast<OpenMesh::PropertyT<py::none>*>(p) != nullptr));
		});
		if (it == this->end(Handle()))
		{
			_out = PropHandle();
			return false;
		}
		else
		{
			_out = PropHandle(idx);
			return true;
		}
		#else
		return Mesh::get_property_handle(_out, _name);
		#endif
	}

	template <class Handle>
	bool py_get_base_property(OpenMesh::BaseProperty *& _prop, int& _idx, const std::string& _name) {
		int idx = -1;
		for (auto it = begin(Handle()); it != end(Handle()); ++it) {
			++idx;
			if (*it != nullptr && (*it)->name() == _name) {
				_prop = *it;
				_idx = idx;
				return true;
			}
		}
		return false;
	}

	template <class Handle, class PropHandle>
	PropHandle py_prop_on_demand(const std::string& _name) {
		PropHandle result;
		if (py_has_property<Handle>(_name)) {
			py_get_property_handle<Handle>(result, _name);
		}
		else {
			Mesh::add_property(result, _name);
		}
		return result;
	}

	typename Mesh::prop_iterator begin(OpenMesh::VertexHandle) { return this->vprops_begin(); }
	typename Mesh::prop_iterator end(OpenMesh::VertexHandle) { return this->vprops_end(); }
	typename Mesh::prop_iterator begin(OpenMesh::HalfedgeHandle) { return this->hprops_begin(); }
	typename Mesh::prop_iterator end(OpenMesh::HalfedgeHandle) { return this->hprops_end(); }
	typename Mesh::prop_iterator begin(OpenMesh::EdgeHandle) { return this->eprops_begin(); }
	typename Mesh::prop_iterator end(OpenMesh::EdgeHandle) { return this->eprops_end(); }
	typename Mesh::prop_iterator begin(OpenMesh::FaceHandle) { return this->fprops_begin(); }
	typename Mesh::prop_iterator end(OpenMesh::FaceHandle) { return this->fprops_end(); }

	template<typename Handle, typename T>
	struct HandleToPropHandle_D
	{};

	template<typename T>
	struct HandleToPropHandle_D<OpenMesh::VertexHandle, T>
	{
		using type = OpenMesh::VPropHandleT<T>;
	};
	template<typename T>
	struct HandleToPropHandle_D<OpenMesh::HalfedgeHandle, T>
	{
		using type = OpenMesh::HPropHandleT<T>;
	};
	template<typename T>
	struct HandleToPropHandle_D<OpenMesh::EdgeHandle, T>
	{
		using type = OpenMesh::EPropHandleT<T>;
	};
	template<typename T>
	struct HandleToPropHandle_D<OpenMesh::FaceHandle, T>
	{
		using type = OpenMesh::FPropHandleT<T>;
	};

	template<typename Handle, typename T>
	using HandleToPropHandle = typename HandleToPropHandle_D<Handle, T>::type;

};

#ifdef OPENMESH_PYTHON_OPENFLIPPER_TYPES
typedef MeshWrapperT<OpenMesh::TriMesh_ArrayKernelT<TriTraits> > TriMesh;
typedef MeshWrapperT<OpenMesh::PolyMesh_ArrayKernelT<PolyTraits> > PolyMesh;
#else
typedef MeshWrapperT<OpenMesh::TriMesh_ArrayKernelT<MeshTraits> > TriMesh;
typedef MeshWrapperT<OpenMesh::PolyMesh_ArrayKernelT<MeshTraits> > PolyMesh;
#endif

} // namespace OpenMeshPython

#endif
