#ifndef OPENMESH_PYTHON_TYPECONVERSIONS_HH
#define OPENMESH_PYTHON_TYPECONVERSIONS_HH

#include <OpenMesh/Core/Geometry/VectorT.hh>
#include <pybind11/pybind11.h>

namespace py = pybind11;

namespace OpenMeshPython {


template<class dtype, int n>
OpenMesh::VectorT<dtype, n> convert_to_vector(py::object _obj) {
	py::array_t<dtype> arr(_obj);
	if (arr.size() != n) {
		std::stringstream msg;
		msg << "could not convert array with size " << arr.size();
		msg << " to " << n << "-dimensional vector";
		PyErr_SetString(PyExc_ValueError, msg.str().c_str());
		throw py::error_already_set();
	}
	return OpenMesh::VectorT<dtype, n>(arr.data());
}

template<class T>
T convert_to_cpp(py::object _obj) {
	return _obj.cast<T>();
}

template<>
inline OpenMesh::VectorT<float, 2> convert_to_cpp(py::object _obj) {
	return convert_to_vector<float, 2>(_obj);
}

template<>
inline OpenMesh::VectorT<float, 3> convert_to_cpp(py::object _obj) {
	return convert_to_vector<float, 3>(_obj);
}

template<>
inline OpenMesh::VectorT<float, 4> convert_to_cpp(py::object _obj) {
	return convert_to_vector<float, 4>(_obj);
}

template<>
inline OpenMesh::VectorT<double, 2> convert_to_cpp(py::object _obj) {
	return convert_to_vector<double, 2>(_obj);
}

template<>
inline OpenMesh::VectorT<double, 3> convert_to_cpp(py::object _obj) {
	return convert_to_vector<double, 3>(_obj);
}

template<>
inline OpenMesh::VectorT<double, 4> convert_to_cpp(py::object _obj) {
	return convert_to_vector<double, 4>(_obj);
}

inline py::object convert_to_python(bool _val) {
	return py::bool_(_val);
}

inline py::object convert_to_python(int _val) {
	return py::int_(_val);
}

inline py::object convert_to_python(float _val) {
	return py::float_(_val);
}

inline py::object convert_to_python(double _val) {
	return py::float_(_val);
}

/**
 * Converts OpenMesh vectors to numpy arrays.
 *
 * The returned array contains a copy of the vector's underlying data.
 *
 * @tparam vector A Vector type.
 * @param _vec The vector to be converted.
 */
template<class Vector>
inline py::array_t<typename Vector::value_type> vec2numpy(const Vector& _vec) {
	typedef typename Vector::value_type dtype;
	dtype *data = new dtype[_vec.size()];
	std::copy_n(_vec.data(), _vec.size(), data);
	py::capsule base = free_when_done(data);
	return py::array_t<dtype>({_vec.size()}, {sizeof(dtype)}, data, base);
}

/**
 * Converts OpenMesh vectors to numpy arrays.
 *
 * The returned array references the vector's underlying data, i.e. changes
 * made to the returned array affect the original mesh.
 *
 * @tparam Mesh A Mesh type.
 * @tparam vector A Vector type.
 *
 * @param _mesh The mesh that owns the vector's underlying memory. In order
 * to avaoid dangling pointers, the lifetime of this mesh is tied to the
 * lifetime of the returned numpy array.
 * @param _vec The vector to be converted.
 */
template<class Mesh, class Vector>
inline py::array_t<typename Vector::value_type> vec2numpy(Mesh& _mesh, Vector& _vec, size_t _n = 1) {
	typedef typename Vector::value_type dtype;
	std::vector<size_t> shape;
	std::vector<size_t> strides;
	if (_n == 1) {
		shape = {_vec.size()};
		strides = {sizeof(dtype)};
	}
	else {
		shape = {_n, _vec.size()};
		strides = {_vec.size() * sizeof(dtype), sizeof(dtype)};
	}
	return py::array_t<dtype>(shape, strides, _vec.data(), py::cast(_mesh));
}

template<class Mesh>
inline py::array_t<float> flt2numpy(Mesh& _mesh, const float& _flt, size_t _n = 1) {
	return py::array_t<float>({_n}, {sizeof(float)}, &_flt, py::cast(_mesh));
}

template<class Mesh>
inline py::array_t<double> flt2numpy(Mesh& _mesh, const double& _flt, size_t _n = 1) {
	return py::array_t<double>({_n}, {sizeof(double)}, &_flt, py::cast(_mesh));
}

} // namespace OpenMeshPython

#endif
